"""Runtime information for a squad (group of characters)"""

from datetime import timedelta, datetime, timezone
import pandas as pd
from smarterbombing.aggregrator.damage_graph_aggregator import DamageGraphAggregator
from smarterbombing.aggregrator.site_statistics_aggregator import SiteStatisticsAggregator
from smarterbombing.configuration import get_log_directory
from smarterbombing.logs.log_locator import get_latest_log_files
from smarterbombing.logs.log_reader import LogReader

def _read_events(reader: LogReader):
    return reader.read_log_events()

def _flatten(in_list):
    return [item for ll in in_list for item in ll]

class Squad:
    """
    Squad - a group of characters

    """

    def __init__(self, name, characters, config):
        self.name = name
        self.characters = characters
        self.config = config

        self.readers = []
        self.aggregators = {}

        self.filter_subjects = config['ignore_list']

    def initialize_aggregators(self):
        """
        Initialize data aggregators for the squad

        """

        dps_average_seconds = self.config['dps_average_seconds']
        realtime_graph_minutes = self.config['live_graph_minutes']
        site_boundary_minimum_seconds = self.config['site_boundary_minimum_seconds']

        outgoing_hostile_damage = DamageGraphAggregator(
            timedelta(minutes=realtime_graph_minutes), dps_average_seconds,
            only_outgoing=True, only_hostile=True)

        outgoing_friendly_damage = DamageGraphAggregator(
            timedelta(minutes=realtime_graph_minutes), dps_average_seconds,
            only_outgoing=True, only_friendly=True)

        incoming_hostile_damage = DamageGraphAggregator(
            timedelta(minutes=realtime_graph_minutes), dps_average_seconds,
            only_incoming=True, only_hostile=True)

        site_statistics = SiteStatisticsAggregator(site_boundary_minimum_seconds)

        self.aggregators['outgoing_hostile_damage'] = outgoing_hostile_damage
        self.aggregators['outgoing_friendly_damage'] = outgoing_friendly_damage
        self.aggregators['incoming_hostile_damage'] = incoming_hostile_damage
        self.aggregators['site_statistics'] = site_statistics

    def initialize_readers(self, follow_logs: bool = True):
        """
        Find and open log files of characters in the squad

        :param follow_logs: should the open logs seek to end

        """

        for reader in self.readers:
            reader.close()
        self.readers = []

        log_directory = get_log_directory(self.config)
        recent_log_files = get_latest_log_files(log_directory)

        for log_file in recent_log_files:
            if log_file['character'] in self.characters:
                reader = LogReader(log_file, self.characters, self.filter_subjects)
                reader.open(follow_logs)

                self.readers.append(reader)

    def read_and_aggregate_data(self):
        """
        Read available data from open logs and aggregate the results

        """

        events = _flatten(map(_read_events, self.readers))

        for _, aggregator in self.aggregators.items():
            aggregator.append_events(events)

        current_time = datetime.now(timezone.utc)
        for _, aggregator in self.aggregators.items():
            aggregator.aggregate(current_time)

    def get_open_files(self):
        """
        Get open files

        :returns: dataframe of open file information

        """

        if len(self.readers) > 0:
            open_files = pd.DataFrame(map(lambda reader: reader.log_file, self.readers))
        else:
            open_files = pd.DataFrame(columns=['filename', 'created_at', 'path', 'character'])

        return open_files

    def get_outgoing_hostile_damage_graph(self):
        """
        Get outgoing hostile damage graph

        :returns: DataFrame of damage to hostiles

        """

        return self.aggregators['outgoing_hostile_damage'].graph_data

    def get_incoming_hostile_damage_graph(self):
        """
        Get incoming hostile damage graph

        :returns: DataFrame of damage from hostiles

        """

        return self.aggregators['incoming_hostile_damage'].graph_data

    def get_compound_site_statistics(self):
        """
        Get compound site statistics

        :returns: DataFrame of compound site statistics

        """

        return self.aggregators['site_statistics'].compound



"""Site statistics aggregator"""
from datetime import datetime, timedelta
import pandas as pd

from smarterbombing.aggregrator.aggregator import Aggregator

_INPUT_COLUMNS = [
    'timestamp',
    'message_type',
    'damage',
    'direction',
    'subject',
    'what',
    'quality',
    'character',
    'friendly_fire',
]

def _safe_division(numerator, denominator):
    return numerator / denominator if denominator else 0.0

class SiteStatisticsAggregator(Aggregator):
    """
    SiteStatisticsAggregator: aggregates combat log events into statistics about sites

    """

    def __init__(self, minimum_gap_seconds: int):
        Aggregator.__init__(self)

        self.minimum_gap_seconds = minimum_gap_seconds

        self.previous_downtime = timedelta(0)

        # Sites done previously
        self.history = pd.DataFrame([])

        # Site in progress (according to minimum gap)
        self.current = pd.DataFrame([])

        # Combined site averages and totals
        self.compound = {
            'total_time': timedelta(0),
            'total_downtime': timedelta(0),
            'total_effective_time': timedelta(0),
            'sites_per_hour': 0,
            'average_downtime': timedelta(0),
            'average_time': timedelta(0),
            'average_effective_time': timedelta(0),
            'average_damage': 0.0,
            'total_damage': 0,
            'damage_per_second': 0.0,
            'time_efficiency': 0.0,
        }

    def _compound_site_statistics(self, data: pd.DataFrame):
        total_time = data.iloc[-1]['end_time'] - data.iloc[0]['start_time']

        total_seconds = total_time.total_seconds()
        total_hours = total_seconds / 3600

        total_downtime = data['downtime'].sum()

        average_downtime = data['downtime'].mean()
        average_site_time = data['duration'].mean()

        average_effective_time = data['effective_duration'].mean()
        effective_time = data['effective_duration'].sum()

        effective_seconds = effective_time.total_seconds()
        time_efficiency = _safe_division(effective_seconds, total_seconds)

        total_site_damage = data['damage'].sum()
        average_site_damage = data['damage'].mean()

        return {
            'total_time': total_time,
            'total_downtime': total_downtime,
            'total_effective_time': effective_time,
            'sites_per_hour': _safe_division(len(data.index),total_hours),
            'average_downtime': average_downtime,
            'average_time': average_site_time,
            'average_effective_time': average_effective_time,
            'average_damage': average_site_damage,
            'total_damage': total_site_damage,
            'damage_per_second': _safe_division(total_site_damage, effective_seconds),
            'time_efficiency': time_efficiency
        }

    def aggregate(self, end_at: datetime) -> pd.DataFrame:
        """
        Aggregate the data appended until now.

        :param prune_old_events: remove old events no longer needed for aggregation
        :param end_at: until what date and time data should be aggregated

        :returns: DataFrame with the site statistics data
        """

        end_at = end_at.replace(tzinfo=None)

        if len(self.events) == 0:
            return pd.concat([self.history, self.current])

        data = pd.DataFrame(self.events)
        data = data[data['direction'] == 'to']
        data = data[data['timestamp'] <= end_at]

        data = data.sort_values(by='timestamp')
        data = data.reset_index()

        # Calculate event deltas
        data['delta_time'] = data['timestamp'].diff()

        # Assign site index based on time
        data['new_site'] = data['delta_time'] > timedelta(seconds=self.minimum_gap_seconds)
        data['site_index'] = data['new_site'].cumsum()

        # Remove old events
        self.events = data[
            data['site_index'] == data['site_index'].max()
        ][_INPUT_COLUMNS].to_dict('records')

        data_no_friendly = data[~data['friendly_fire']]

        site_downtime = data[data['new_site']][['site_index', 'delta_time']]\
            .set_index('site_index')\
            .rename(columns={'delta_time': 'downtime'})

        if not site_downtime.empty:
            self.previous_downtime = site_downtime.iloc[-1]['downtime']

        if site_downtime.empty and not self.history.empty:
            site_downtime = pd.DataFrame([{
                'downtime': self.previous_downtime
            }])

        site_groups = data.groupby('site_index')
        start_times = site_groups['timestamp'].first()
        end_times = site_groups['timestamp'].last()

        site_groups = data_no_friendly.groupby('site_index')
        effective_start_times = site_groups['timestamp'].first()
        effective_end_times = site_groups['timestamp'].last()

        site_damage = data_no_friendly.pivot(columns='site_index', values='damage').sum()
        site_hits = data_no_friendly.pivot(columns='site_index', values='subject').count()

        durations = end_times - start_times
        effective_durations = effective_end_times - effective_start_times

        result = pd.DataFrame({
            'effective_start_time': effective_start_times,
            'effective_end_time': effective_end_times,
            'start_time': start_times,
            'end_time': end_times,
            'duration': durations,
            'effective_duration': effective_durations,
            'damage': site_damage,
            'hits': site_hits,
        })

        result = result.combine_first(site_downtime).fillna(timedelta(0))

        self.history = pd.concat([self.history, result.iloc[:-1]], ignore_index=True)
        self.current = result.iloc[-1:]

        if len(self.history.index) > 0:
            self.compound = self._compound_site_statistics(pd.concat([self.history, self.current]))

        return result



"""Main application entrypoint"""
from argparse import ArgumentParser
import sys

import smarterbombing.webui.ui as ui

parser = ArgumentParser(
    prog='smarterbombing',
    description='Parse Eve Online combat logs and display statistics.',
    epilog='Show your support by sending ISK in-game to Ageliten.')

MODE_HELP = 'How to run smarterbombing -- webui (default)'
PORT_HELP = 'Which port to use for hosting the webui.'

parser.add_argument(
    '--mode',
    default='webui',
    required=False,
    help=MODE_HELP)

parser.add_argument(
    '--port',
    default=42069,
    required=False,
    help=PORT_HELP
)

args = parser.parse_args()

if args.mode == 'webui':
    print('Starting WebUI...')
    ui.run_webui(args.port)
else:
    print(f'unrecognized mode: {args.mode}')
    sys.exit(1)



"""Squads interface"""
import gradio as gr

from smarterbombing import configuration
from smarterbombing.webui.ui_squad import render_squad

def render_squads(sb_ui: gr.Blocks, config: dict):
    """
    Render squads user interface

    :param sb_ui: the Blocks user interface object
    :param config: the configuration

    """

    squads = configuration.get_squads(config)

    for squad_index in range(len(squads)):
        squad = config['squads'][squad_index]
        render_squad(sb_ui, config, squad)



"""Damage graph aggregator"""
from datetime import datetime, timedelta
import pandas as pd

from smarterbombing.aggregrator.aggregator import Aggregator

def _average_dps_per_character_melt(data: pd.DataFrame) -> pd.DataFrame:
    return data.reset_index().melt(id_vars='timestamp', value_name='damage')

def _filter_by_datetime(data: pd.DataFrame, start_date: datetime, end_date: datetime):
    return data[(data['timestamp'] >= start_date) & (data['timestamp'] <= end_date)]

def _generate_1s_timeseries_dataframe(start_date: datetime, end_date: datetime) -> pd.DataFrame:
    start_date = start_date.replace(microsecond=0)
    end_date = end_date.replace(microsecond=0)

    timerange = pd.date_range(start_date, end_date, freq='S', unit='s')

    return timerange.to_frame(index=False, name='timestamp')

class DamageGraphAggregator(Aggregator):
    """
    DamageGraphAggregator - aggregates combat log events into average damage over time data
    """

    def __init__(self,
                 graph_time_window: timedelta,
                 average_over_seconds: int,
                 only_incoming: bool = False,
                 only_outgoing: bool = False,
                 only_friendly: bool = False,
                 only_hostile: bool = False,
                ):
        Aggregator.__init__(self)

        self.graph_time_window = graph_time_window
        self.graph_data = pd.DataFrame(columns=['timestamp', 'character', 'damage'])

        self.average_over_seconds = average_over_seconds

        self.only_incoming = only_incoming
        self.only_outgoing = only_outgoing
        self.only_friendly = only_friendly
        self.only_hostile = only_hostile

    def aggregate(self, end_at: datetime) -> pd.DataFrame:
        """
        Aggregate the data appended until now.

        :param end_at: until what date and time data should be aggregated

        :returns: DataFrame with the damage per second graph data

        """

        end_at = end_at.replace(tzinfo=None)

        start_at = end_at - (
            self.graph_time_window + timedelta(seconds = self.average_over_seconds)
        )
        graph_start_at = end_at - self.graph_time_window

        if len(self.events) > 0:
            data = pd.DataFrame(self.events)
        else:
            data = pd.DataFrame(
                columns=['timestamp', 'character', 'friendly_fire', 'direction', 'damage']
            )

        data = _filter_by_datetime(data, start_at, end_at)

        if len(data.index) > 0:
            if self.only_incoming:
                data = data[data['direction'] == 'from']

            if self.only_outgoing:
                data = data[data['direction'] == 'to']

            if self.only_friendly:
                data = data[data['friendly_fire']]

            if self.only_hostile:
                data = data[~data['friendly_fire']]

        # Remove old events
        self.events = data.to_dict('records')

        data = data[['timestamp', 'character', 'damage']]

        characters = data['character'].unique()

        fixed_window = pd.concat([
            _generate_1s_timeseries_dataframe(start_at, end_at), pd.DataFrame(columns=characters),
        ]).fillna(0.0).set_index('timestamp').rename_axis('character', axis='columns')

        data = data.groupby(['timestamp', 'character']).sum().reset_index()
        data = data.pivot(
            index='timestamp',
            columns='character',
            values='damage'
        ).fillna(0.0)

        data = data.combine_first(fixed_window)
        data = data.assign(Total=data.sum(1))
        data = data.resample('1S').asfreq(fill_value=0.0)

        if self.average_over_seconds > 0:
            data = data.rolling(timedelta(seconds=self.average_over_seconds)).mean()

        # NOTE(axel): Cut off at graph starting point after averaging, this prevents values
        # drifting as they get close to the start graph time window.
        data = data[data.index >= graph_start_at]

        self.graph_data = _average_dps_per_character_melt(data)

        return data



"""Functions for parsing Eve Online combat log messages"""
import re
from datetime import datetime

REGEX_REMOVE_LOG_FORMATTING = re.compile('<.*?>')

REGEX_MATCH_COMBAT_LOG = re.compile(
    r'^\[\ (.*?)\ \]\ \((combat)\) (\d*?) (to|from) (.*?)(?!\[.*?\]\(.*?\)) - ')

REGEX_REMOVE_SHIP_TYPE = re.compile(
    r'\[.*?\]\(.*?\)'
)

def strip_log_formatting(log_message: str) -> str:
    """
    Remove any formatting tags from log message.

    :param log_message: a log message
    :returns: the same log message without any formatting tags

    """

    return re.sub(REGEX_REMOVE_LOG_FORMATTING, '', log_message)

def parse_log_timestamp(log_datetime: str) -> datetime:
    """
    Parse log date/time representation to datetime object.

    :param log_datetime: a datetime string, ex. 2023.10.05 15:30:00
    :returns: a datetime object
    :raises ValueError: log_datetime doesn't match expected format

    """

    try:
        return datetime(int(log_datetime[:4]), int(log_datetime[5:7]), int(log_datetime[8:10]),
             int(log_datetime[11:13]), int(log_datetime[14:16]), int(log_datetime[17:]))
    except ValueError as inner:
        raise ValueError(f'Invalid date string: {log_datetime}') from inner

def parse_combat_log_line(log_line):
    """
    Parse one Eve Online combat log line.

    The result will contain the following keys:
    - timestamp (datetime)
    - message_type (str)    -- combat
    - damage (int)          -- eg. 345
    - direction (str)       -- to, from
    - subject (str)         -- eg. Centus Tyrant
    - what (str)            -- eg. Large EMP Smartbomb II
    - quality (str)         -- eg. Hits

    :param log_line: the raw log line
    :returns: a dict containing the parsed log or None if parsing failed

    """

    log_line = re.sub(REGEX_REMOVE_LOG_FORMATTING, '', log_line)

    match = re.match(REGEX_MATCH_COMBAT_LOG, log_line)
    if match is None:
        return None

    (
        timestamp,
        message_type,
        damage,
        direction,
        subject,
    ) = match.groups()

    subject = re.sub(REGEX_REMOVE_SHIP_TYPE, '', subject)

    timestamp = parse_log_timestamp(timestamp)

    log_line = log_line[match.end():]

    parts = log_line.split('-')

    if len(parts) > 1:
        [what, quality] = parts
    else:
        [what, quality] = ['', parts[0]]

    return {
        'timestamp': timestamp,
        'message_type': message_type,
        'damage': int(damage),
        'direction': direction,
        'subject': subject,
        'what': what.strip(),
        'quality': quality.strip(),
    }



"""Interface for collecting and aggregating data"""
from datetime import datetime
import pandas as pd

class Aggregator(object):
    """
    Aggregator: interface for collecting and aggregating data

    """

    def __init__(self):
        self.events = []

    def append_events(self, events: list):
        """
        Append events to aggregator.

        :param events: list of events

        """
        self.events.extend(events)

    def aggregate(self, end_at: datetime) -> pd.DataFrame:
        """
        Aggregate the data appended until now.

        :param end_at: until what date and time data should be aggregated

        :returns: DataFrame with the damage per second graph data

        """
